import re
from django.db import models
from django.conf import settings
from django.core.exceptions import ValidationError
from django.db.models.signals import post_save
from django.dispatch import receiver

import jsonfield

import logging
logger = logging.getLogger(__name__)


def set_on_delete(collector, field, sub_objs, using):
    default_group_hosts = []
    existing_hosts = []
    for host in HostGroup.objects.get(pk=1).hosts.all():
        default_group_hosts.append(host.hostname)
    for host_group in collector.data[HostGroup]:
        for host in host_group.hosts.all():
            if host.hostname in default_group_hosts:
                existing_hosts.append(host.hostname)
    if existing_hosts:
        raise ValidationError("""Hosts {} exists in default group.
        Please set another group or delete this hosts.""".format(", ".join(existing_hosts)), 10001)
    collector.add_field_update(field, 1, sub_objs)
set_on_delete.deconstruct = lambda: ('django.db.models.SET', (1,), {})


class HostGroup(models.Model):
    name = models.CharField("Name", max_length=255, unique=True)
    smart_proxy = models.ForeignKey('SmartProxy', null=False)
    group_fields = jsonfield.JSONField('Group fields', null=True)
    source = models.ForeignKey('HostSource', null=True, on_delete=models.SET_NULL)
    source_params = jsonfield.JSONField(null=True)

    class Meta:
        app_label = "pxetool"

    def __unicode__(self):
        return self.name

    def __getattr__(self, item):
        if type(self.group_fields) == dict and item in self.group_fields:
            return self.group_fields[item]
        raise AttributeError

    @property
    def config_url(self):
        return self.smart_proxy.pxetool_url

    @property
    def pxe_server(self):
        return self.smart_proxy.hostname

    @property
    def pxe_server_tftpboot(self):
        return self.smart_proxy.tftpboot

    @property
    def repos(self):
        item = 'additional_repos'
        if item in self.group_fields:
            return self.group_fields[item]
        else:
            return []

    @property
    def mirror_url(self):
        return self.mirrors

    @property
    def mirror_host(self):
        host_res = settings.MIRROR_REGEX.match(self.mirror_url)
        if host_res:
            return host_res.groupdict().get('hostname')
        return ''

    @property
    def mirror_directory(self):
        host_res = settings.MIRROR_REGEX.match(self.mirror_url)
        if host_res:
            return host_res.groupdict().get('path')
        return ''


class Host(models.Model):
    hostname = models.CharField("Hostname", max_length=255)
    mac = models.CharField("MAC address", max_length=17, unique=True)
    profile = models.ForeignKey('Profile', related_name='host', null=True)
    profile_params = jsonfield.JSONField(null=True)
    installation = models.ForeignKey('Installation', related_name='+', null=True)
    group = models.ForeignKey('HostGroup', related_name='hosts', null=False, on_delete=set_on_delete)
    created = models.DateTimeField("Created", auto_now_add=True)
    source = models.ForeignKey('HostSource', null=True, on_delete=models.SET_NULL)

    def __getattr__(self, item):
        if item not in ['group', 'profile_params'] and not re.match('^_', item):
            sources = [
                self.group,
            ]
            if self.profile_params and \
                    item in self.profile_params:
                return self.profile_params[item]
            for i in sources:
                if hasattr(i, item):
                    return getattr(i, item)
        raise AttributeError("'{}' object has no attribute '{}'".format(
            self.__class__.__name__, item
        ))

    def __unicode__(self):
        return '{} / {}'.format(self.group.name, self.hostname)

    class Meta:
        unique_together = ('hostname', 'group')
        ordering = ('hostname', )
        app_label = "pxetool"

    @classmethod
    def create(cls, args):
        from pxetool.models import form_profile_host
        profile_form = form_profile_host(data=args)
        host = cls(
            hostname=args['hostname'],
            mac=args['mac'],
            group=HostGroup.objects.get(pk=args['group']),
        )
        if profile_form.is_valid():
            host.profile_params = profile_form.cleaned_data
        host.save()
        from pxetool_drivers.models import DriverManager
        driver_opt = {}
        driver_name = args['driver_name']
        driver_opt['driver_name'] = args['driver_name']
        driver_opt['params'] = {}
        for arg, value in args.items():
            arg_regex = re.match(driver_name + '\-(.*)', arg)
            if arg_regex:
                driver_opt['params'][arg_regex.groups()[0]] = value
        driver = DriverManager(host=host, **driver_opt)
        driver.save()
        return host

    @property
    def domain(self):
        return '.'.join(self.hostname.split('.')[1:])

    @property
    def short_hostname(self):
        return self.hostname.split('.')[0]

    @property
    def get_dashed_mac(self):
        return '01-%s' % re.sub(':', '-', self.mac.lower())

    @property
    def status(self):
        return self.installation.status if self.installation else 'unknown'

    @property
    def color(self):
        st = {
            'pending_installation': 'active',
            'booting_to_installer': 'info',
            'installing': 'info',
            'booting_to_os': 'info',
            'configuring': 'warning',
            'done': 'success',
            'error': 'danger',
            'aborted': 'danger',
            'unknown': '',
        }
        return st[self.status if self.status in st else 'unknown']

    @property
    def installing(self):
        if self.installation.status not in ['done', 'error', 'aborted']:
            return self.installation
        return None

    def reboot(self):
        self.driver.power_cycle(host=self)

    def power_status(self):
        return self.driver.power_state()


class HostSource(models.Model):
    name = models.CharField(max_length=255)
    url = models.CharField(max_length=255)
    type = models.CharField(
        max_length=10,
        choices=map(lambda x: (x, x), ('libvirt', 'inventory')),
        default='inventory'
    )
    auth_data = jsonfield.JSONField(null=True)
    enable = models.BooleanField(default=True)

    class Meta:
        app_label = "pxetool"

    def __unicode__(self):
        return "{} (URL: {})".format(self.name, self.url)

    @property
    def x_auth_application(self):
        if 'X-Auth-Application' not in self.auth_data:
            return ''
        return self.auth_data['X-Auth-Application']

    @property
    def x_auth_token(self):
        if 'X-Auth-Token' not in self.auth_data:
            return ''
        return self.auth_data['X-Auth-Token']


@receiver(post_save)
def post_save_host_source(sender, instance, **kwargs):
    if sender not in [HostSource, HostGroup]:
        return
    from pxetool.tasks import fetch_hosts_from_source
    fetch_hosts_from_source()
