import celery.decorators
from celery.utils.log import get_task_logger
from django.db import IntegrityError
from django.conf import settings

import os
import pwd
import errno
import tempfile
import paramiko
import urllib
import requests

logger = get_task_logger(__name__)

# set only because flake8 can't find variables declaration
smart_proxy = profile = installation = host = None


def init_objects_wrapper(func):
    from pxetool.models import SmartProxy, Profile, Installation, Host

    def wrapper(*args, **kwargs):
        func_varnames = func.func_code.co_varnames
        keys = {
            'smart_proxy_pk': {'name': 'smart_proxy', 'model': SmartProxy},
            'profile_pk': {'name': 'profile', 'model': Profile},
            'install_pk': {'name': 'installation', 'model': Installation},
            'host_pk': {'name': 'host', 'model': Host},
        }
        for pk_name in keys.keys():
            pk_val = \
                args[func_varnames.index(pk_name)] \
                if pk_name in func_varnames \
                else kwargs.get(pk_name, None)
            if pk_val:
                func.func_globals[keys[pk_name]['name']] = \
                    keys[pk_name]['model'].objects.get(pk=pk_val)
        func(*args, **kwargs)
    return wrapper


def _init_ssh(smart_proxy, *args, **kwargs):
    pkey = paramiko.RSAKey.from_private_key_file(os.path.join(
        pwd.getpwuid(os.geteuid())[5],
        ".ssh/id_rsa",
    ))
    ssh = paramiko.SSHClient()
    ssh.load_system_host_keys()
    # ToDo: show ssh key on smart-proxy add and save it to model
    #       instead of AutoAddPolicy
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh.connect(
        str(smart_proxy.hostname),
        port=smart_proxy.port,
        username=str(smart_proxy.username),
        pkey=pkey,
    )
    return ssh


@init_objects_wrapper
def _upload_file(smart_proxy_pk, src, dst, *args, **kwargs):
    ssh = _init_ssh(smart_proxy, *args, **kwargs)
    sftp = ssh.open_sftp()
    if not os.path.exists(src):
        logger.error("File {} doesn't exist".format(src))
        return
    # check directory existence
    dst_dirname = os.path.dirname(dst)
    try:
        try:
            f_stat = sftp.lstat(dst_dirname)
        except IOError as e:
            f_stat = None
        if not f_stat:
            sftp.mkdir(dst_dirname)
    except IOError as e:
        logger.error("Unable to create remote directory '{}' on host '{}': {}".format(
            dst_dirname, smart_proxy.hostname, str(e)))
        return
    try:
        if sftp.lstat(dst) and 'overwrite' in kwargs:
            sftp.put(src, dst)
    except IOError as e:
        if e.errno == errno.ENOENT:
            logger.debug("Uploading {} => {}".format(src, dst))
            sftp.put(src, dst)
    if 'owner' in kwargs and 'group' in kwargs:
        try:
            sftp.chown(dst, kwargs['owner'], kwargs['group'])
        except (IOError, OSError) as e:
            logger.warn("Unable to chown file '{}' on host '{}': {}".format(
                dst, smart_proxy.hostname, str(e)))
    if 'mode' in kwargs:
        try:
            sftp.chmod(dst, kwargs['mode'])
        except (IOError, OSError) as e:
            logger.warn("Unable to chmod file '{}' on host '{}': {}".format(
                dst, smart_proxy.hostname, str(e)))


@celery.decorators.task(name="upload_profile_task")
@init_objects_wrapper
def upload_profile_task(smart_proxy_pk, profile_pk, *args, **kwargs):
    for i in ('kernel', 'initrd'):
        if not getattr(profile, i):
            continue
        logger.debug("Uploading {} :: source {}".format(i, getattr(profile, i)))
        filename = getattr(profile, i)
        local_path = os.path.join(settings.PROFILES_FILES, filename)
        remote_path = os.path.join(smart_proxy.tftpboot, filename)
        _upload_file(smart_proxy_pk, local_path, remote_path)


@celery.decorators.task(name="upload_file_task")
def upload_file_task(*args, **kwargs):
    _upload_file(*args, **kwargs)


@celery.decorators.task(name="upload_file_content_task")
@init_objects_wrapper
def upload_file_content_task(smart_proxy_pk, content, dst, *args, **kwargs):
    f = tempfile.NamedTemporaryFile(mode='w', delete=False)
    f.write(content)
    f.close()
    _upload_file(smart_proxy_pk, f.name, dst, *args, **kwargs)
    os.remove(f.name)


@celery.decorators.task(name="update_install_status_task")
@init_objects_wrapper
def update_install_status_task(install_pk, status, *args, **kwargs):
    installation.status = status
    installation.save()


@celery.decorators.task(name="delete_file_task")
@init_objects_wrapper
def delete_file_task(smart_proxy_pk, filename, *args, **kwargs):
    ssh = _init_ssh(smart_proxy, *args, **kwargs)
    sftp = ssh.open_sftp()
    try:
        sftp.remove(filename)
    except IOError as e:
        if e.errno == errno.ENOENT:
            return
        logger.error(e)


@celery.decorators.task(name="power_cycle_task")
@init_objects_wrapper
def power_cycle_task(host_pk, *args, **kwargs):
    host.driver.power_cycle()


def inventory_importer(source, params=None):
    url = source.url
    if type(params) == dict:
        url += "?{}".format(urllib.urlencode(params))
    try:
        response = requests.get(
            url,
            headers={
                'X-Auth-Application': source.x_auth_application,
                'X-Auth-Token': source.x_auth_token,
            },
        )
    except requests.exceptions.RequestException as e:
        logger.error(e.message)
        return {}
    data = response.json()
    inventory_hosts = {}
    for host in data:
        if 'attributes' not in host:
            # skip hosts without attributes
            continue
        inventory_hosts[host['name']] = host['attributes'].get('macaddress_eth0')
    return inventory_hosts


@celery.decorators.task(name="fetch_hosts_from_source")
def fetch_hosts_from_source(*args, **kwargs):
    from .models import Host, HostGroup
    from pxetool_drivers.models import DriverManager
    for hg in HostGroup.objects.filter(source__isnull=False, source__enable=True):
        source = hg.source
        hosts = {}
        if source.type == 'inventory':
            inventory_hosts = inventory_importer(source, hg.source_params)
            for hostname in inventory_hosts:
                hosts[hostname] = {
                    'mac': inventory_hosts[hostname],
                    'driver_name': 'powerless',
                    'driver_params': {},
                }
        elif source.type == 'libvirt':
            from pxetool_drivers.virsh import Virsh
            virsh = Virsh(None, conn_string=source.url)
            for hostname, domain in virsh.list_domains().iteritems():
                hosts[hostname] = {
                    'mac': domain['interfaces'][0],
                    'driver_name': 'virsh',
                    'driver_params': {'conn_string': source.url},
                }
        existing_hosts = {h.hostname: h for h in Host.objects.filter(group=hg, source=source)}
        for hostname, host in hosts.iteritems():
            if hostname in existing_hosts:
                del existing_hosts[hostname]
                continue
            try:
                new_host = Host.objects.create(
                    hostname=hostname,
                    mac=host['mac'],
                    group=hg,
                    source=source,
                )
                DriverManager.objects.create(
                    driver_name=host['driver_name'],
                    params=host['driver_params'],
                    host=new_host,
                )
            except IntegrityError:
                pass
        for host in existing_hosts.values():
            host.delete()
