import os
import pwd

from django.test import TestCase
from django.db import IntegrityError
from django.contrib.auth.models import User
from .models import Host, HostGroup, SmartProxy
from .models import Installation


class SetUp(object):
    sm = None
    hosts = []
    host_groups = []
    user = None

    def setUp(self):
        self.user = User.objects.create_superuser(
            email='user@host.local',
            username='user',
            password='user',
        )
        self.sm = SmartProxy.objects.create(
            hostname='localhost',
            username=pwd.getpwuid(os.getuid()).pw_name,
        )
        self.host_groups = [
            HostGroup.objects.create(name='test_group{}'.format(i), smart_proxy=self.sm)
            for i in range(1, 3)
        ]
        self.hosts = [
            Host.objects.create(
                hostname='hg1-host{}.test.local'.format(i),
                mac='01-01-01-01-01-{:02d}'.format(i),
                group=self.host_groups[0],
            )
            for i in range(1, 3)
        ]
        self.hosts.extend([
            Host.objects.create(
                hostname='hg2-host{}.test.local'.format(i),
                mac='01-01-01-01-02-{:02d}'.format(i),
                group=self.host_groups[1],
            )
            for i in range(1, 3)
        ])


class HostsTestCase(SetUp, TestCase):
    def test_class_to_string(self):
        """Test correct __unicode__ function"""
        self.assertEqual(str(self.host_groups[0]), self.host_groups[0].name)
        self.assertEqual(
            str(self.hosts[0]),
            '{} / {}'.format(self.hosts[0].group.name, self.hosts[0].hostname)
        )

    def test_same_host_different_group(self):
        src_host = Host.objects.filter(group=self.host_groups[0]).first()
        Host.objects.create(
            hostname=src_host.hostname,
            mac='00-01-02-03-04-06',
            group=self.host_groups[1]
        )

    def test_same_host_same_group(self):
        with self.assertRaises(IntegrityError):
            Host.objects.create(hostname='hg1-host1.test.local', mac='00-01-02-03-04-07', group=self.host_groups[0])

    def test_delete_second_group(self):
        """check on delete group all hosts moving to first group"""
        hosts_count = Host.objects.count()
        self.host_groups[1].delete()
        self.assertEqual(Host.objects.count(), hosts_count)

    def test_color_property(self):
        st = {
            'pending_installation': 'active',
            'booting_to_installer': 'info',
            'installing': 'info',
            'booting_to_os': 'info',
            'configuring': 'warning',
            'done': 'success',
            'error': 'danger',
            'aborted': 'danger',
            'unknown': '',
            '__nonexistent_item___': '',
        }
        host = Host()
        self.assertEqual(host.status, 'unknown')
        host = self.hosts[0]
        installation = Installation(host=host, owner=self.user)
        installation.save()
        host.installation = installation
        host.save()
        for k, v in st.iteritems():
            installation.status = k
            installation.save()
            self.assertEqual(host.color, v)
